import os
import pathlib

from gi.repository import Gtk

from .translation import gettext as _
from .image_formats import get_supported_input_format_mimetypes
from .image_formats import get_supported_input_format_exts


def _build_file_filters():
    image_file_filter = Gtk.FileFilter()
    image_file_filter.set_name(_("Image Files"))
    if os.name == "nt":
        for ext in get_supported_input_format_exts():
            image_file_filter.add_pattern("*%s" % ext)
    else:
        for mimetype in get_supported_input_format_mimetypes():
            image_file_filter.add_mime_type(mimetype)

    any_file_filter = Gtk.FileFilter()
    any_file_filter.set_name(_("All Files"))
    any_file_filter.add_pattern("*")

    return [image_file_filter, any_file_filter]


def _build_gtk_file_chooser_open(parent=None):
    file_chooser_dialog = Gtk.FileChooserDialog(
        title=_("Open Images..."),
        parent=parent,
    )
    file_chooser_dialog.add_buttons(
        Gtk.STOCK_CANCEL,
        Gtk.ResponseType.CANCEL,
        Gtk.STOCK_OPEN,
        Gtk.ResponseType.ACCEPT,
    )
    file_chooser_dialog.set_select_multiple(True)

    for filter_ in _build_file_filters():
        file_chooser_dialog.add_filter(filter_)

    return file_chooser_dialog


def _build_gtk_file_chooser_save(filename=None, parent=None):
    file_chooser_dialog = Gtk.FileChooserDialog(
        title=_("Save Image As..."),
        parent=parent,
    )
    file_chooser_dialog.add_buttons(
        Gtk.STOCK_CANCEL,
        Gtk.ResponseType.CANCEL,
        Gtk.STOCK_SAVE,
        Gtk.ResponseType.ACCEPT,
    )
    file_chooser_dialog.set_action(Gtk.FileChooserAction.SAVE)

    for filter_ in _build_file_filters():
        file_chooser_dialog.add_filter(filter_)

    if filename:
        file_chooser_dialog.set_filename(filename)
        file_chooser_dialog.set_current_name(pathlib.Path(filename).name)

    return file_chooser_dialog


def _build_win32_file_chooser_open():
    file_chooser_dialog = Gtk.FileChooserNative(
        title=_("Open Images..."),
    )
    file_chooser_dialog.set_select_multiple(True)

    for filter_ in _build_file_filters():
        file_chooser_dialog.add_filter(filter_)

    return file_chooser_dialog


def _build_win32_file_chooser_save(filename=None):
    file_chooser_dialog = Gtk.FileChooserNative(
        title=_("Save Image As..."),
    )
    file_chooser_dialog.set_action(Gtk.FileChooserAction.SAVE)

    for filter_ in _build_file_filters():
        file_chooser_dialog.add_filter(filter_)

    if filename:
        file_chooser_dialog.set_filename(filename)
        file_chooser_dialog.set_current_name(pathlib.Path(filename).name)

    return file_chooser_dialog


def open_file_chooser_open_file(parent=None):
    if os.name == "nt":
        file_chooser_dialog = _build_win32_file_chooser_open()
    else:
        file_chooser_dialog = _build_gtk_file_chooser_open(parent=parent)

    response = file_chooser_dialog.run()

    filenames = []
    if response == Gtk.ResponseType.ACCEPT:
        filenames = file_chooser_dialog.get_filenames()

    file_chooser_dialog.destroy()

    return filenames


def open_file_chooser_save_file(filename=None, parent=None):
    if os.name == "nt":
        file_chooser_dialog = _build_win32_file_chooser_save(filename=filename)
    else:
        file_chooser_dialog = _build_gtk_file_chooser_save(
            filename=filename, parent=parent
        )

    response = file_chooser_dialog.run()

    filename = None
    if response == Gtk.ResponseType.ACCEPT:
        filename = file_chooser_dialog.get_filename()

    file_chooser_dialog.destroy()

    return filename



import os
from pathlib import Path
import ctypes
from configparser import ConfigParser

from . import APPLICATION_ID


OUTPUT_PATTERN_NEXT_TO_FILE = "next-to-file"
OUTPUT_PATTERN_SUBFOLDER = "subfolder"
OUTPUT_PATTERN_CUSTOM = "custom"
DEFAULT_OUTPUT_PATTERNS = {
    "next-to-file": "{FILENAME}.opti.{EXT}",
    "subfolder": "optimized/{FILENAME}.{EXT}",
}

DEFAULT = {
    "optimization": {
        "threads": 2,
    },
    "interface": {
        "gtk-theme-name": "default",
        "gtk-application-prefer-dark-theme": False,
    },
    "output": {
        "active-pattern": OUTPUT_PATTERN_NEXT_TO_FILE,
        "custom-pattern": "optimized/{EXT}/example_{FILENAME}.{EXT}",
    },
}


def get_unix_xdg_config_home():
    """Returns the path to the user's config directory (XDG_CONFIG_HOME) on
    Unix platforms.

    :rtype: pathlib.Path
    """
    if "XDG_CONFIG_HOME" in os.environ and os.environ["XDG_CONFIG_HOME"]:
        config_dir = Path(os.environ["XDG_CONFIG_HOME"])
    else:
        config_dir = Path("~/.config")
    return config_dir.expanduser().resolve().absolute()


def get_win_user_data_dir():
    """Returns the user's data path on Windows platform.

    :rtype: pathlib.Path
    """
    if "APPDATA" in os.environ and os.environ["APPDATA"]:
        user_data_dir = Path(os.environ["APPDATA"])
    else:
        CSIDL_APPDATA = 26
        MAX_PATH = 2048

        buff = ctypes.create_unicode_buffer(MAX_PATH)
        ctypes.windll.shell32.SHGetFolderPathW(
            None, CSIDL_APPDATA, None, 0, buff
        )

        buff2 = ctypes.create_unicode_buffer(MAX_PATH)
        ctypes.windll.kernel32.GetShortPathNameW(buff.value, buff2, MAX_PATH)

        user_data_dir = Path(buff2.value)

    return user_data_dir.expanduser().resolve().absolute()


def get_config_file_path():
    """Returns the path to the application's config file.

    :rtype: pathlib.Path
    """
    if os.name == "posix":
        return get_unix_xdg_config_home() / APPLICATION_ID / "config.ini"
    elif os.name == "nt":
        return get_win_user_data_dir() / APPLICATION_ID / "config.ini"


def get_config():
    """Get the application configuration.

    :rtype: configparser.ConfigParser
    """
    config = ConfigParser()

    # Load defaults
    config.read_dict(DEFAULT)

    # Load user configuration
    config_path = get_config_file_path()
    if config_path.exists():
        config.read(config_path)

    return config


def save_config(config):
    """Save the given configuration.

    :param configparser.ConfigParser config: The configuration to save.
    """
    config_path = get_config_file_path()

    if not config_path.parent.exists():
        config_path.parent.mkdir(parents=True)

    with open(config_path, "w") as config_file:
        config.write(config_file)



# File from: https://gist.github.com/flozz/2560bcced07abde5713d3eeae839bf60

"""
This module contains helpers to deal with GTK 3 themes in Python / PyGObject.


Dependencies
------------

* Python 3
* PyGObject
* GLib, Gio and GTK with introspection files


Run the examples
----------------

::

    python3 gtk_themes_helpers.py


Author
------

* Fabien LOISON (https://github.com/flozz)


License
-------

::

    .           DO WHAT THE FUCK YOU WANT TO PUBLIC LICENSE
                        Version 2, December 2004

     Copyright (C) 2004 Sam Hocevar <sam@hocevar.net>

     Everyone is permitted to copy and distribute verbatim or modified
     copies of this license document, and changing it is allowed as long
     as the name is changed.

                DO WHAT THE FUCK YOU WANT TO PUBLIC LICENSE
       TERMS AND CONDITIONS FOR COPYING, DISTRIBUTION AND MODIFICATION

      0. You just DO WHAT THE FUCK YOU WANT TO.


See also
--------

The following code was inspired by GTK Inspector and Inkscape sources:

* https://gitlab.gnome.org/GNOME/gtk/-/blob/352898ae/gtk/inspector/visual.c#L536
* https://gitlab.com/inkscape/inkscape/-/blob/53780867/src/ui/themes.cpp#L69
"""


from pathlib import Path

from gi.repository import Gtk, Gio, GLib


def get_gtk_theme_name():
    """Get the name of the currently used GTK theme.

    :rtype: str
    """
    settings = Gtk.Settings.get_default()
    return settings.get_property("gtk-theme-name")


def set_gtk_theme_name(gtk_theme_name):
    """Set the GTK theme to use.

    :param str gtk_theme_name: The name of the theme to use (e.g.
                               ``"Adwaita"``).
    """
    settings = Gtk.Settings.get_default()
    settings.set_property("gtk-theme-name", gtk_theme_name)


def get_gtk_application_prefer_dark_theme():
    """Returns ``True`` if GTK is currently using the dark variant of the
    theme.

    :rtype: bool
    """
    settings = Gtk.Settings.get_default()
    return settings.get_property("gtk-application-prefer-dark-theme")


def set_gtk_application_prefer_dark_theme(use_dark_theme):
    """Defines whether the dark variant of the GTK theme should be used or
    not.

    :param bool use_dark_theme: If ``True`` the dark variant of the theme will
                                be used (if available).
    """
    settings = Gtk.Settings.get_default()
    settings.set_property("gtk-application-prefer-dark-theme", use_dark_theme)


def list_gtk_themes():
    """Lists all available GTK theme.

    :rtype: [str]
    """
    builtin_themes = [
        theme[:-1]
        for theme in Gio.resources_enumerate_children(
            "/org/gtk/libgtk/theme", Gio.ResourceFlags.NONE
        )
    ]

    theme_search_dirs = [
        Path(data_dir) / "themes" for data_dir in GLib.get_system_data_dirs()
    ]
    theme_search_dirs.append(Path(GLib.get_user_data_dir()) / "themes")
    theme_search_dirs.append(Path(GLib.get_home_dir()) / ".themes")
    fs_themes = []
    for theme_search_dir in theme_search_dirs:
        if not theme_search_dir.exists():
            continue

        for theme_dir in theme_search_dir.iterdir():
            if (
                not (theme_dir / "gtk-3.0" / "gtk.css").exists()
                and not (theme_dir / "gtk-3.0" / "gtk-dark.css").exists()
                and not (theme_dir / "gtk-3.20" / "gtk.css").exists()
                and not (theme_dir / "gtk-3.20" / "gtk-dark.css").exists()
            ):
                continue
            fs_themes.append(theme_dir.as_posix().split("/")[-1])

    return sorted(set(builtin_themes + fs_themes))


if __name__ == "__main__":
    print("Available themes: %s" % ", ".join(list_gtk_themes()))
    print("Current theme: %s" % get_gtk_theme_name())
    print("Use dark variant: %s" % get_gtk_application_prefer_dark_theme())



import sys
import argparse

from .application import YogaImageOptimizerApplication
from . import VERSION


def main(args=sys.argv):
    # Basic CLI using argparse (simpler than using GTK Application CLI)
    cli_parser = argparse.ArgumentParser()
    cli_parser.add_argument(
        "-v",
        "--version",
        action="version",
        version=VERSION,
    )
    cli_parser.add_argument(
        "image",
        nargs="*",
        help="Input image",
    )
    cli_parser.parse_args(args[1:])

    # GTK Application
    app = YogaImageOptimizerApplication()
    app.run(args)


if __name__ == "__main__":
    main()



import os

from gi.repository import Gtk, GdkPixbuf

from . import APPLICATION_NAME, APPLICATION_ID
from . import data_helpers
from . import gtk_themes_helpers
from .translation import gtk_builder_translation_hack
from .translation import gettext as _
from .config import save_config


class SettingsWindow(Gtk.Window):
    def __init__(self, config, parent_window=None):
        Gtk.Window.__init__(
            self,
            title="%s - %s" % (_("Settings"), APPLICATION_NAME),
            icon=GdkPixbuf.Pixbuf.new_from_file(
                data_helpers.find_data_path("images/icon_64.png")
            ),
            default_width=450,
            modal=True,
        )
        self.set_transient_for(parent_window)
        self.set_position(Gtk.WindowPosition.CENTER_ON_PARENT)

        self._config = config

        self._builder = Gtk.Builder()
        self._builder.set_translation_domain(APPLICATION_ID)
        self._builder.add_from_file(
            data_helpers.find_data_path("ui/settings-window.glade")
        )
        self._builder.connect_signals(self)

        content = self._builder.get_object("settings_window_content")
        self.add(content)

        self._prepare_theme_combobox()
        self.update_interface()

        self.connect("destroy", self._on_settings_windows_destroyed)

        # HACK: Translate the UI on Windows
        if os.name == "nt":
            gtk_builder_translation_hack(self._builder)

    def destroy(self, *args):
        Gtk.Window.destroy(self)

    def update_interface(self):
        # Optimization / Threads
        threads_adjustment = self._builder.get_object("threads_adjustment")
        threads_adjustment.set_value(
            self._config.getint("optimization", "threads")
        )

        # Interface / Theme
        if (
            gtk_themes_helpers.get_gtk_theme_name()
            in gtk_themes_helpers.list_gtk_themes()
        ):
            theme_combobox = self._builder.get_object("theme_combobox")
            theme_combobox.set_active(
                gtk_themes_helpers.list_gtk_themes().index(
                    gtk_themes_helpers.get_gtk_theme_name()
                )
            )

        # Interface / Prefer dark theme
        prefer_dark_theme_switch = self._builder.get_object(
            "prefer_dark_theme_switch"
        )
        prefer_dark_theme_switch.set_state(
            self._config.getboolean(
                "interface", "gtk-application-prefer-dark-theme"
            )
        )

        # Output file / Output files location
        output_pattern_radiobuttons = {
            "next-to-file": self._builder.get_object(
                "output_pattern_next_to_file_radiobutton"
            ),
            "subfolder": self._builder.get_object(
                "output_pattern_subfolder_radiobutton"
            ),
            "custom": self._builder.get_object(
                "output_pattern_custom_radiobutton"
            ),
        }
        output_pattern_radiobuttons[
            self._config.get("output", "active-pattern")
        ].set_active(True)

        output_pattern_custom_entry = self._builder.get_object(
            "output_pattern_custom_entry"
        )
        output_pattern_custom_entry.set_text(
            self._config.get("output", "custom-pattern")
        )

        output_pattern_custom_entry.set_sensitive(
            self._config.get("output", "active-pattern") == "custom"
        )

    def _prepare_theme_combobox(self):
        theme_combobox = self._builder.get_object("theme_combobox")

        for theme in gtk_themes_helpers.list_gtk_themes():
            theme_combobox.append_text(theme)

    def _on_threads_adjustment_value_changed(self, adjustment):
        self._config.set(
            "optimization", "threads", str(int(adjustment.get_value()))
        )

    def _on_theme_combobox_changed(self, widget):
        gtk_theme = gtk_themes_helpers.list_gtk_themes()[widget.get_active()]
        self._config.set("interface", "gtk-theme-name", gtk_theme)
        gtk_themes_helpers.set_gtk_theme_name(gtk_theme)

    def _on_prefer_dark_theme_switch_state_setted(self, widget, state):
        self._config.set(
            "interface", "gtk-application-prefer-dark-theme", str(state)
        )
        gtk_themes_helpers.set_gtk_application_prefer_dark_theme(state)

    def _on_output_pattern_next_to_file_radiobutton_toggled(self, widget):
        if not widget.get_active():
            return
        self._config.set("output", "active-pattern", "next-to-file")
        output_pattern_custom_entry = self._builder.get_object(
            "output_pattern_custom_entry"
        )
        output_pattern_custom_entry.set_sensitive(False)

    def _on_output_pattern_subfolder_radiobutton_toggled(self, widget):
        if not widget.get_active():
            return
        self._config.set("output", "active-pattern", "subfolder")
        output_pattern_custom_entry = self._builder.get_object(
            "output_pattern_custom_entry"
        )
        output_pattern_custom_entry.set_sensitive(False)

    def _on_output_pattern_custom_radiobutton_toggled(self, widget):
        if not widget.get_active():
            return
        self._config.set("output", "active-pattern", "custom")
        output_pattern_custom_entry = self._builder.get_object(
            "output_pattern_custom_entry"
        )
        output_pattern_custom_entry.set_sensitive(True)

    def _on_output_pattern_custom_entry_changed(self, widget):
        self._config.set("output", "custom-pattern", widget.get_text())

    def _on_settings_windows_destroyed(self, widget):
        save_config(self._config)



import os
from pathlib import Path

import yoga.image
from gi.repository import Gtk
from gi.repository import GLib
from gi.repository import Gio

from . import APPLICATION_ID
from . import helpers
from . import config
from . import gtk_themes_helpers
from .image_formats import IMAGES_FORMATS
from .image_formats import find_file_format
from .main_window import MainWindow
from .about_dialog import AboutDialog
from .settings_window import SettingsWindow
from .image_store import ImageStore
from .file_chooser import open_file_chooser_open_file
from .stoppable_process_pool_executor import StoppableProcessPoolExecutor


class YogaImageOptimizerApplication(Gtk.Application):

    STATE_MANAGE_IMAGES = "manage"
    STATE_OPTIMIZE = "optimize"

    def __init__(self):
        Gtk.Application.__init__(
            self,
            application_id=APPLICATION_ID,
            flags=Gio.ApplicationFlags.HANDLES_OPEN,
        )

        self.current_state = None
        self.image_store = ImageStore()
        self.config = config.get_config()

        self._main_window = None
        self._settings_window = None
        self._executor = None
        self._futures = []

    def do_startup(self):
        Gtk.Application.do_startup(self)

        # Action: app.about
        action = Gio.SimpleAction.new("about", None)
        action.connect("activate", lambda a, p: self.about())
        self.add_action(action)

        # Action: app.settings
        action = Gio.SimpleAction.new("settings", None)
        action.connect("activate", lambda a, p: self.settings())
        self.add_action(action)

        # Action: app.quit
        action = Gio.SimpleAction.new("quit", None)
        action.connect("activate", lambda a, p: self.quit())
        self.add_action(action)
        self.set_accels_for_action("app.quit", ["<Ctrl>Q", "<Ctrl>W"])

        # Action: app.optimize
        action = Gio.SimpleAction.new("optimize", None)
        action.connect("activate", lambda a, p: self.optimize())
        self.add_action(action)

        # Action: app.stop-optimization
        action = Gio.SimpleAction.new("stop-optimization", None)
        action.connect("activate", lambda a, p: self.stop_optimization())
        self.add_action(action)

        # Action: app.add-image
        action = Gio.SimpleAction.new("add-image", GLib.VariantType("s"))
        action.connect("activate", lambda a, p: self.add_image(p.get_string()))
        self.add_action(action)

        # Action: app.clear-images
        action = Gio.SimpleAction.new("clear-images", None)
        action.connect("activate", lambda a, p: self.clear_images())
        self.add_action(action)

        # Action: app.open-file
        action = Gio.SimpleAction.new("open-file", None)
        action.connect("activate", lambda a, p: self.open_file())
        self.add_action(action)
        self.set_accels_for_action("app.open-file", ["<Ctrl>O"])

        # Apply GTK theme
        if self.config.get("interface", "gtk-theme-name") != "default":
            gtk_themes_helpers.set_gtk_theme_name(
                self.config.get("interface", "gtk-theme-name")
            )
        gtk_themes_helpers.set_gtk_application_prefer_dark_theme(
            self.config.getboolean(
                "interface", "gtk-application-prefer-dark-theme"
            )
        )

    def do_activate(self):
        if not self._main_window:
            self._main_window = MainWindow(self)
            self.switch_state(self.STATE_MANAGE_IMAGES)

        self._main_window.show()
        self._main_window.present()

    def do_open(self, files, file_count, hint):
        self.do_activate()

        if self.current_state == self.STATE_OPTIMIZE:
            # TODO display a message to inform the user we cannot add files now
            return

        for file_ in files:
            self.add_image(file_.get_path())

    def about(self):
        about_dialog = AboutDialog(parent=self._main_window)
        about_dialog.run()
        about_dialog.destroy()

    def settings(self):
        def _on_settings_window_destroyed(*args):
            self._settings_window = None

        if not self._settings_window:
            self._settings_window = SettingsWindow(
                self.config, parent_window=self._main_window
            )
            self._settings_window.connect(
                "destroy", _on_settings_window_destroyed
            )
            self._settings_window.show_all()
        self._settings_window.present()

    def quit(self):
        self.stop_optimization()
        Gtk.Application.quit(self)

    def switch_state(self, state):
        self.current_state = state
        self._main_window.switch_state(state)

    def add_image(self, path):
        input_path = Path(path).resolve()
        input_size = input_path.stat().st_size
        input_format = find_file_format(input_path)

        if input_format is None:
            print("W: File ignored (unsupported format): %s" % path)
            return

        output_format = input_format

        # Default to JPEG if the input format is not supported as output format
        if not IMAGES_FORMATS[input_format]["output"]:
            output_format = "jpeg"

        active_pattern = self.config.get("output", "active-pattern")
        if active_pattern in config.DEFAULT_OUTPUT_PATTERNS:
            output_pattern = config.DEFAULT_OUTPUT_PATTERNS[active_pattern]
        elif active_pattern == config.OUTPUT_PATTERN_CUSTOM:
            output_pattern = self.config.get("output", "custom-pattern")
        else:
            output_pattern = config.DEFAULT_OUTPUT_PATTERNS[
                config.OUTPUT_PATTERN_NEXT_TO_FILE
            ]

        image = helpers.open_image_from_path(str(input_path))

        self.image_store.append(
            input_file=str(input_path),
            input_size=input_size,
            output_size=0,
            input_format=input_format,
            output_format=output_format,
            preview=helpers.preview_gdk_pixbuf_from_image(image),
            image_width=image.width,
            image_height=image.height,
            use_output_pattern=True,
            output_pattern=output_pattern,
        )

        image.close()

    def clear_images(self):
        self.image_store.clear()

    def open_file(self):
        filenames = open_file_chooser_open_file(parent=self._main_window)
        for filename in filenames:
            self.add_image(filename)

    def optimize(self):
        self.switch_state(self.STATE_OPTIMIZE)

        self._executor = StoppableProcessPoolExecutor(
            max_workers=self.config.getint("optimization", "threads")
        )
        self._futures = []

        for row in self.image_store.get_all():
            if row["status"] == ImageStore.STATUS_DONE:
                # Skip already optimized images
                self._futures.append(None)
                continue
            if not Path(row["output_file"]).parent.is_dir():
                Path(row["output_file"]).parent.mkdir(parents=True)
            self._futures.append(
                self._executor.submit(
                    yoga.image.optimize,
                    row["input_file"],
                    row["output_file"],
                    {
                        "output_format": row["output_format"].lower(),
                        "jpeg_quality": row["jpeg_quality"] / 100,
                        "webp_quality": row["webp_quality"] / 100,
                        "png_slow_optimization": row["png_slow_optimization"],
                        "resize": [
                            row["resize_width"],
                            row["resize_height"],
                        ]
                        if row["resize_enabled"]
                        else "orig",
                    },
                )
            )

        self._update_optimization_status()

    def stop_optimization(self):
        if self.current_state != self.STATE_OPTIMIZE:
            return

        self._executor.shutdown(wait=False)
        self.switch_state(self.STATE_MANAGE_IMAGES)

        for i in range(self.image_store.length):
            image = self.image_store.get(i)
            if image["status"] in [
                self.image_store.STATUS_PENDING,
                self.image_store.STATUS_IN_PROGRESS,
            ]:
                self.image_store.update(
                    i, status=self.image_store.STATUS_CANCELED
                )

    def _update_optimization_status(self):
        if self.current_state != self.STATE_OPTIMIZE:
            return

        is_running = False

        for i in range(len(self._futures)):
            future = self._futures[i]

            if future is None:
                continue

            if future.running():
                self.image_store.update(
                    i,
                    status=self.image_store.STATUS_IN_PROGRESS,
                    output_size=0,
                )
                is_running = True
            elif future.done():
                image_data = self.image_store.get(i)
                output_size = os.stat(image_data["output_file"]).st_size

                self.image_store.update(
                    i,
                    status=self.image_store.STATUS_DONE,
                    output_size=output_size,
                )
            else:
                self.image_store.update(
                    i,
                    status=self.image_store.STATUS_PENDING,
                    output_size=0,
                )
                is_running = True

        if is_running:
            GLib.timeout_add_seconds(0.1, self._update_optimization_status)
        else:
            self.stop_optimization()



import os
import locale
import gettext

from . import APPLICATION_ID
from . import data_helpers


if "LANG" not in os.environ:
    language, encoding = locale.getlocale()
    os.environ["LANG"] = language

translation = gettext.translation(
    APPLICATION_ID,
    localedir=data_helpers.find_data_path("locales"),
    fallback=True,
)

if hasattr(locale, "bindtextdomain"):
    locale.bindtextdomain(
        APPLICATION_ID, data_helpers.find_data_path("locales")
    )
else:
    print("W: Unable to bind text domaine")

gettext = translation.gettext
format_string = locale.format_string


def gtk_builder_translation_hack(builder):
    """Hack to allow translating the UI on Windows.

    :param Gtk.Builder builder: The builder to hack...
    """
    translatable_properties = [
        "label",
        "text",
        "title",
        "tooltip-text",
    ]
    for widget in builder.get_objects():
        widget_properties = [prop.name for prop in widget.list_properties()]
        for translatable_property in translatable_properties:
            if translatable_property not in widget_properties:
                continue
            text = widget.get_property(translatable_property)
            if not text:
                continue
            widget.set_property(translatable_property, gettext(text))



import os
from pathlib import Path

from gi.repository import Gtk, Gdk, Gio, GdkPixbuf

from . import APPLICATION_NAME, APPLICATION_ID
from . import helpers
from . import data_helpers
from . import config
from .image_formats import get_supported_output_format_ids
from .image_formats import get_supported_output_format_names
from .translation import gtk_builder_translation_hack
from .translation import gettext as _
from .custom_pattern_dialog import CustomPatternDialog
from .file_chooser import open_file_chooser_save_file


class MainWindow(Gtk.ApplicationWindow):
    def __init__(self, app):
        Gtk.ApplicationWindow.__init__(
            self,
            application=app,
            title=APPLICATION_NAME,
            icon=GdkPixbuf.Pixbuf.new_from_file(
                data_helpers.find_data_path("images/icon_64.png")
            ),
            default_width=1000,
            default_height=600,
            resizable=True,
        )

        self._updating_interface = False

        self._builder = Gtk.Builder()
        self._builder.set_translation_domain(APPLICATION_ID)
        self._builder.add_from_file(
            data_helpers.find_data_path("ui/main-window.glade")
        )
        self._builder.connect_signals(self)

        header = self._builder.get_object("main_window_header")
        self.set_titlebar(header)

        content = self._builder.get_object("main_window_content")
        self.add(content)

        self._prepare_treeview()
        self._prepare_format_combobox()

        self.connect("destroy", self._on_main_window_destroyed)

        # Drag & drop files
        self.drag_dest_set(
            Gtk.DestDefaults.ALL,
            [
                Gtk.TargetEntry.new(
                    "text/uri-list",
                    Gtk.TargetFlags.OTHER_APP,
                    0,
                )
            ],
            Gdk.DragAction.COPY,
        )
        self.connect("drag-data-received", self._on_drag_data_received)

        # Action: win.remove-selected-image
        action = Gio.SimpleAction.new("remove-selected-images", None)
        action.connect(
            "activate", self._on_remove_selected_images_action_activated
        )
        self.add_action(action)

        images_treeview = self._builder.get_object("images_treeview")
        self.set_accels_for_action_while_widget_focused(
            "win.remove-selected-images",
            ["Delete"],
            images_treeview,
        )

        # HACK: Translate the UI on Windows
        if os.name == "nt":
            gtk_builder_translation_hack(self._builder)

    def set_accels_for_action_while_widget_focused(
        self, detailed_action_name, accels, widget
    ):
        action_group_name, action_name = detailed_action_name.split(".")
        action_group = self.get_action_group(action_group_name)

        accel_group = Gtk.AccelGroup()

        for key, mods in [Gtk.accelerator_parse(accel) for accel in accels]:
            accel_group.connect(
                key,
                mods,
                0,
                # Activate the action only if the widget is focused
                lambda *_: widget.has_focus()
                and action_group.activate_action(action_name),
            )

        self.add_accel_group(accel_group)

    def switch_state(self, state):
        app = self.get_application()

        # fmt: off
        if state == app.STATE_MANAGE_IMAGES:
            self._builder.get_object("add_image_button").set_sensitive(True)
            self._builder.get_object("remove_image_button").set_sensitive(True)
            self._builder.get_object("clear_images_button").set_sensitive(True)
            self._builder.get_object("optimize_button").show()
            self._builder.get_object("stop_optimization_button").hide()
            self._builder.get_object("output_image_options").set_sensitive(True)
            self._builder.get_object("jpeg_options").set_sensitive(True)
            self._builder.get_object("webp_options").set_sensitive(True)
            self._builder.get_object("png_options").set_sensitive(True)
        elif state == app.STATE_OPTIMIZE:
            self._builder.get_object("add_image_button").set_sensitive(False)
            self._builder.get_object("remove_image_button").set_sensitive(False)
            self._builder.get_object("clear_images_button").set_sensitive(False)
            self._builder.get_object("optimize_button").hide()
            self._builder.get_object("stop_optimization_button").show()
            self._builder.get_object("output_image_options").set_sensitive(False)
            self._builder.get_object("jpeg_options").set_sensitive(False)
            self._builder.get_object("webp_options").set_sensitive(False)
            self._builder.get_object("png_options").set_sensitive(False)
        # fmt: on

    def update_interface(self):
        self._updating_interface = True

        app = self.get_application()
        output_image_options = self._builder.get_object("output_image_options")
        jpeg_options = self._builder.get_object("jpeg_options")
        webp_options = self._builder.get_object("webp_options")
        png_options = self._builder.get_object("png_options")

        def _have_all_same_values(iters, property_):
            value = app.image_store.get(iters[0])[property_]
            for iter_ in iters:
                if app.image_store.get(iter_)[property_] != value:
                    return False
            return True

        def _avg(iters, property_):
            return sum(
                [app.image_store.get(iter_)[property_] for iter_ in iters]
            ) // len(iters)

        # Reset output options visibility (hide everything)
        output_image_options.hide()
        jpeg_options.hide()
        webp_options.hide()
        png_options.hide()

        # Get selected images
        iters = self.get_selected_image_iters()

        # No image selected, stop here
        if len(iters) == 0:
            return

        if _have_all_same_values(iters, "output_format"):
            output_format = app.image_store.get(iters[0])["output_format"]
        else:
            output_format = None

        # [Output image] Update and show output image options
        output_format_combobox = self._builder.get_object(
            "output_format_combobox"
        )
        if output_format is not None:
            output_format_combobox.set_active(
                get_supported_output_format_ids().index(output_format)
            )
        else:
            output_format_combobox.set_active(-1)

        resize_checkbutton = self._builder.get_object("resize_checkbutton")
        if _have_all_same_values(iters, "resize_enabled"):
            resize_checkbutton.set_active(
                app.image_store.get(iters[0])["resize_enabled"]
            )
            resize_checkbutton.set_inconsistent(False)
        else:
            resize_checkbutton.set_active(False)
            resize_checkbutton.set_inconsistent(True)

        resize_width_adjustment = self._builder.get_object(
            "resize_width_adjustment"
        )
        resize_width_adjustment.set_value(
            max(
                [app.image_store.get(iter_)["resize_width"] for iter_ in iters]
            )
        )

        resize_width_spinbutton = self._builder.get_object(
            "resize_width_spinbutton"
        )
        resize_width_spinbutton.set_sensitive(resize_checkbutton.get_active())

        resize_height_adjustment = self._builder.get_object(
            "resize_height_adjustment"
        )
        resize_height_adjustment.set_value(
            max(
                [
                    app.image_store.get(iter_)["resize_height"]
                    for iter_ in iters
                ]
            )
        )

        resize_height_spinbutton = self._builder.get_object(
            "resize_height_spinbutton"
        )
        resize_height_spinbutton.set_sensitive(resize_checkbutton.get_active())

        resize_reset_button = self._builder.get_object("resize_reset_button")
        if len(iters) > 1:
            resize_reset_button.set_visible(True)
        elif (
            app.image_store.get(iters[0])["resize_width"]
            != app.image_store.get(iters[0])["image_width"]
            or app.image_store.get(iters[0])["resize_height"]
            != app.image_store.get(iters[0])["image_height"]
        ):
            resize_reset_button.set_visible(True)
        else:
            resize_reset_button.set_visible(False)

        output_file_entry = self._builder.get_object("output_file_entry")
        if len(iters) == 1:
            output_file = app.image_store.get(iters[0])["output_file"]
            output_file_entry.set_text(output_file)
            output_file_entry.set_sensitive(True)
        else:
            output_file_entry.set_text(_("— Multiple files selected —"))
            output_file_entry.set_sensitive(False)

        output_image_options.show()

        output_path_browse_modelbutton = self._builder.get_object(
            "output_path_browse_modelbutton"
        )
        output_path_browse_modelbutton.set_sensitive(len(iters) == 1)

        # [JPEG] Update and show jpeg options
        if output_format == "jpeg":
            jpeg_quality_adjustment = self._builder.get_object(
                "jpeg_quality_adjustment"
            )
            jpeg_quality_adjustment.set_value(_avg(iters, "jpeg_quality"))
            jpeg_options.show()
        # [WebP] Update and show webp options
        elif output_format == "webp":
            webp_quality_adjustment = self._builder.get_object(
                "webp_quality_adjustment"
            )
            webp_quality_adjustment.set_value(_avg(iters, "webp_quality"))
            webp_options.show()
        # [PNG] Update and show png options
        elif output_format == "png":
            png_slow_optimization_checkbutton = self._builder.get_object(
                "png_slow_optimization_checkbutton"
            )
            if _have_all_same_values(iters, "png_slow_optimization"):
                png_slow_optimization_checkbutton.set_active(
                    app.image_store.get(iters[0])["png_slow_optimization"]
                )
                png_slow_optimization_checkbutton.set_inconsistent(False)
            else:
                png_slow_optimization_checkbutton.set_active(False)
                png_slow_optimization_checkbutton.set_inconsistent(True)
            png_options.show()

        self._updating_interface = False

    def remove_selected_images(self):
        app = self.get_application()
        if not app.current_state == app.STATE_MANAGE_IMAGES:
            return

        iters = self.get_selected_image_iters()
        for iter_ in iters:
            app = self.get_application()
            app.image_store.remove(iter_)

    def get_selected_image_iters(self):
        treeview_images = self._builder.get_object("images_treeview")
        selection = treeview_images.get_selection()
        model = treeview_images.get_model()

        _, tree_paths = selection.get_selected_rows()
        return [model.get_iter(tree_path) for tree_path in tree_paths]

    def _prepare_treeview(self):
        app = self.get_application()

        DISPLAYED_FIELDS = [
            "status_display",
            "preview",
            "input_file_display",
            "input_size_display",
            "separator",
            "output_file_display",
            "output_format_display",
            "output_size_display",
        ]

        treeview_images = self._builder.get_object("images_treeview")
        treeview_images.set_model(app.image_store.gtk_list_store)

        for field_name in DISPLAYED_FIELDS:
            field_type = app.image_store.FIELDS[field_name]["type"]

            if field_type is str:
                treeview_images.append_column(
                    Gtk.TreeViewColumn(
                        app.image_store.FIELDS[field_name]["label"],
                        Gtk.CellRendererText(),
                        text=app.image_store.FIELDS[field_name]["id"],
                    ),
                )
            elif field_type is GdkPixbuf.Pixbuf:
                pixbuf = Gtk.CellRendererPixbuf()
                column = Gtk.TreeViewColumn(
                    app.image_store.FIELDS[field_name]["label"],
                    pixbuf,
                )
                column.add_attribute(
                    pixbuf,
                    "pixbuf",
                    app.image_store.FIELDS[field_name]["id"],
                )
                treeview_images.append_column(column)
            else:
                raise TypeError(
                    "Unsupported field type '%s'" % str(field_type)
                )

        # Enable multi-selection
        selection = treeview_images.get_selection()
        selection.set_mode(Gtk.SelectionMode.MULTIPLE)

    def _prepare_format_combobox(self):
        output_format_combobox = self._builder.get_object(
            "output_format_combobox"
        )

        for output_format in get_supported_output_format_names():
            output_format_combobox.append_text(output_format)

    def _on_main_window_destroyed(self, widget):
        app = self.get_application()
        app.quit()

    def _on_drag_data_received(
        self, widget, drag_context, x, y, data, info, time
    ):
        app = self.get_application()

        def _add_path(paths):
            for path in paths:
                if path.is_dir():
                    _add_path(p for p in path.glob("**/*") if p.is_file())
                elif path.is_file():
                    app.add_image(path)
                    Gtk.main_iteration_do(True)

        for uri in data.get_uris():
            path = Path(helpers.gvfs_uri_to_local_path(uri))
            _add_path([path])

    def _on_image_treeview_selection_changed(self, selection):
        self.update_interface()

    def _on_output_format_combobox_changed(self, combobox):
        if self._updating_interface:
            return

        app = self.get_application()
        output_format_combobox = self._builder.get_object(
            "output_format_combobox"
        )

        # No format selected
        if output_format_combobox.get_active() == -1:
            return

        output_format = get_supported_output_format_ids()[
            output_format_combobox.get_active()
        ]

        iters = self.get_selected_image_iters()
        for iter_ in iters:
            app.image_store.update(iter_, output_format=output_format)
            app.image_store.reset_status(iter_)

        self.update_interface()

    def _on_resize_checkbutton_toggled(self, checkbutton):
        if self._updating_interface:
            return

        app = self.get_application()
        iters = self.get_selected_image_iters()

        for iter_ in iters:
            app.image_store.update(
                iter_, resize_enabled=checkbutton.get_active()
            )
            app.image_store.reset_status(iter_)

        self.update_interface()

    def _on_resize_width_adjustment_value_changed(self, adjustment):
        if self._updating_interface:
            return

        app = self.get_application()
        iters = self.get_selected_image_iters()

        for iter_ in iters:
            app.image_store.update(
                iter_, resize_width=round(adjustment.get_value())
            )
            app.image_store.reset_status(iter_)

        self.update_interface()

    def _on_resize_height_adjustment_value_changed(self, adjustment):
        if self._updating_interface:
            return

        app = self.get_application()
        iters = self.get_selected_image_iters()

        for iter_ in iters:
            app.image_store.update(
                iter_, resize_height=round(adjustment.get_value())
            )
            app.image_store.reset_status(iter_)

        self.update_interface()

    def _on_resize_reset_button_clicked(self, widget):
        if self._updating_interface:
            return

        app = self.get_application()
        iters = self.get_selected_image_iters()

        for iter_ in iters:
            app.image_store.update(
                iter_,
                resize_width=app.image_store.get(iter_)["image_width"],
                resize_height=app.image_store.get(iter_)["image_height"],
            )
            app.image_store.reset_status(iter_)

        self.update_interface()

    def _on_output_file_entry_changed(self, entry):
        if self._updating_interface:
            return

        app = self.get_application()
        iters = self.get_selected_image_iters()

        if len(iters) != 1:
            return

        output_file = Path(entry.get_text())
        app.image_store.update(
            iters[0],
            output_file=str(output_file.resolve()),
            use_output_pattern=False,
        )
        app.image_store.reset_status(iters[0])

    def _on_output_pattern_next_to_file_modelbutton_clicked(self, widget):
        app = self.get_application()
        iters = self.get_selected_image_iters()

        for iter_ in iters:
            app.image_store.update(
                iter_,
                output_pattern=config.DEFAULT_OUTPUT_PATTERNS["next-to-file"],
                use_output_pattern=True,
            )
            app.image_store.reset_status(iter_)

        if len(iters) == 1:
            output_file_entry = self._builder.get_object("output_file_entry")
            output_file_entry.set_text(
                app.image_store.get(iters[0])["output_file"]
            )

    def _on_output_pattern_subfolder_modelbutton_clicked(self, widget):
        app = self.get_application()
        iters = self.get_selected_image_iters()

        for iter_ in iters:
            app.image_store.update(
                iter_,
                output_pattern=config.DEFAULT_OUTPUT_PATTERNS["subfolder"],
                use_output_pattern=True,
            )
            app.image_store.reset_status(iter_)

        if len(iters) == 1:
            output_file_entry = self._builder.get_object("output_file_entry")
            output_file_entry.set_text(
                app.image_store.get(iters[0])["output_file"]
            )

    def _on_output_pattern_custom_modelbutton_clicked(self, widget):
        app = self.get_application()

        custom_pattern_dialog = CustomPatternDialog(
            initial_pattern=app.config.get("output", "custom-pattern"),
            parent_window=self,
        )
        custom_pattern = custom_pattern_dialog.run()

        if not custom_pattern:
            return

        iters = self.get_selected_image_iters()

        for iter_ in iters:
            app.image_store.update(
                iter_,
                output_pattern=custom_pattern,
                use_output_pattern=True,
            )
            app.image_store.reset_status(iter_)

        if len(iters) == 1:
            output_file_entry = self._builder.get_object("output_file_entry")
            output_file_entry.set_text(
                app.image_store.get(iters[0])["output_file"]
            )

    def _on_output_path_browse_modelbutton_clicked(self, widget):
        app = self.get_application()
        iter_ = self.get_selected_image_iters()[0]

        filename = open_file_chooser_save_file(
            filename=app.image_store.get(iter_)["output_file"],
            parent=self,
        )

        if filename:
            app.image_store.update(
                iter_,
                output_file=filename,
                use_output_pattern=False,
            )
            output_file_entry = self._builder.get_object("output_file_entry")
            output_file_entry.set_text(
                app.image_store.get(iter_)["output_file"]
            )

    def _on_jpeg_quality_adjustement_value_changed(self, adjustment):
        if self._updating_interface:
            return

        app = self.get_application()
        iters = self.get_selected_image_iters()

        for iter_ in iters:
            app.image_store.update(
                iter_, jpeg_quality=round(adjustment.get_value())
            )
            app.image_store.reset_status(iter_)

    def _on_webp_quality_adjustement_value_changed(self, adjustment):
        if self._updating_interface:
            return

        app = self.get_application()
        iters = self.get_selected_image_iters()

        for iter_ in iters:
            app.image_store.update(
                iter_, webp_quality=round(adjustment.get_value())
            )
            app.image_store.reset_status(iter_)

    def _on_png_slow_optimization_checkbutton_toggled(self, checkbutton):
        if self._updating_interface:
            return

        app = self.get_application()
        iters = self.get_selected_image_iters()

        for iter_ in iters:
            app.image_store.update(
                iter_, png_slow_optimization=checkbutton.get_active()
            )
            app.image_store.reset_status(iter_)

        self.update_interface()

    def _on_remove_selected_images_action_activated(self, action, param):
        self.remove_selected_images()
